#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include "bool.h"
#include "cdf.h"
#include "intmatrix.h"
#include "entry.h"
#include "main.h"

static const int INFINITY = 1000000;

static int Num_positives;
static Entry_T Entries[MAXPOSITIVES];

#if TEXT
static char Seq[MAXSEQLEN];
static char Name[MAXNAMELEN+1];
#endif

static int
read_matrices (FILE *fp, int **calib_matrix, int *score_matrix,
	       char *accession, char *description, unsigned long *N) {
  int c, pos = 0, h, i, j, numLines, averageint;
  int value, m = 0;
  char aa;
  double average;

  if ((c = fgetc(fp)) != EOF) {
    i = 0;
    accession[i++] = c;

    while ((c = fgetc(fp)) != '\n') {
      accession[i++] = c;
    }
    
    accession[i] = '\0';

    i = 0;
    
    while ((c = fgetc(fp)) != '\n') {
      description[i++] = c;
    }
    
    description[i] = '\0';
    
    fscanf(fp, "%lu", N);
    while (fgetc(fp) != '\n') ;

    fscanf(fp, "%d", &numLines);
    while (fgetc(fp) != '\n') ;

    for (h = 0; h < numLines; h++) {
      fscanf(fp,"%d",&value);
      average = 0.0;
      calib_matrix[pos][0] = value;
      score_matrix[m++] = value;
      average += (double) value * AA_acd_freq(0);
      
      for (i = 1; i < AA_abc_length; ++i) {
	aa = AA_abc_aa(i);
	fscanf(fp,"%d",&value);
	
	if ((j = AA_acd_index(aa)) != -1) {
	  calib_matrix[pos][j] = value;
	  score_matrix[m++] = value;
	  average += (double) value * AA_acd_freq(j);
	}
      }
      
      averageint = (int) rint(average);
      score_matrix[m++] = averageint;
      while (fgetc(fp) != '\n') ;

      ++pos;
    }

    while (fgetc(fp) != '\n') ;
  }

  return (pos);
}

void
sfloor_compute (int *sfloor, int **calib_matrix, int width,
	       int threshold) {
  int column_max, i, j;
  
  sfloor[width-1] = threshold;
  for (i = width-1; i >= 1; --i) {
    column_max = -INFINITY;
    for (j = 0; j < NAT_AAS; ++j) {
      if (calib_matrix[i][j] > column_max) {
	column_max = calib_matrix[i][j];
      }
    }
    sfloor[i-1] = sfloor[i] - column_max;
  }
}

#if TEXT
static void
read_name (FILE *fp, char *name) {
  int i = 0;
  int c = 0;

  while (!feof(fp) && (i < MAXNAMELEN) && 
	 ((c = fgetc(fp)) != '\n')) {
    name[i++] = c;
  }
  if (feof(fp)) {
    fprintf(stderr, "File ended while reading name of sequence.\n");
    name[i] = '\0';
  } else if (i >= MAXNAMELEN) {
    while (fgetc(fp) != '\n') ;
    name[i-1] = '\0';
  } else {
    name[i] = '\0';
  }
}


static bool
read_sequence (FILE *fp, char *seq, int *length, char *name) {
  int i;
  int c = 0, offset;
  
  while (!feof(fp) && fgetc(fp) != '>') ;
  if (feof(fp)) {
    *length = 0;
    return FALSE;
  } else {
    read_name(fp,name);
  }

  i = 0;
  while (!feof(fp) && (i < MAXSEQLEN) &&
	 ((c = fgetc(fp)) != '>')) {
    if (isalpha(c)) {
      offset = (char) (AA_acdx_index(c) * sizeof(MATENTRY_T));
      memcpy(seq+i, &offset, sizeof(char));
      ++i;
    }
  }
  if (feof(fp)) {
    *length = i;
    return TRUE;
  } else if (i >= MAXSEQLEN) {
    fprintf(stderr,
	    "Sequence %s exceeds max length of %d\n -- truncated\n",
	    name,MAXSEQLEN);
    while (fgetc(fp) != '\n') ;
    *length = i;
    return TRUE;
  } else {
    ungetc(c,fp);
    *length = i;
    return TRUE;
  }
}
#endif


#if TEXT
static void
hit_translate (char *hit, char *seq, int start, int width) {
  int i, j;

  for (i = start, j = 0; j < width; ++i, ++j) {
    hit[j] = AA_acdx_aa(seq[i]/sizeof(MATENTRY_T));
  }
  hit[j] = '\0';
}
#else
static void
hit_translate (char *hit, short *sequence_int, int start, int width) {
  int i, j;

  for (i = start, j = 0; j < width; ++i, ++j) {
    hit[j] = AA_acdx_aa((int) sequence_int[i]);
  }
  hit[j] = '\0';
}
#endif


#if TEXT
static void
score_sequences (FILE *fp, char *score_matrix, int threshold, 
		 int *sfloor, double *G, int min_score, int width) {
  register int score;
  register char *m, *s;
  int *sfloor_ptr;
  int length, start, end, pos;
  double significance;
  int namelen;
  char *hit = (char *) calloc(width+1,sizeof(char));

  Name[MAXNAMELEN] = '\0';
  Num_positives = 0;

  while (read_sequence(fp,Seq,&length,Name)) {
    if (length >= width) {
      end = length - width;
      for (start = 0; start <= end; ++start) {
	s = Seq + start;
	score = * (MATENTRY_T *) (score_matrix + *s++);
	for (m = score_matrix + SCOREMAT_SKIP, pos = 1,
	       sfloor_ptr = sfloor;
	     (pos < width) && (score >= *sfloor_ptr++);
	     m += SCOREMAT_SKIP, ++pos) {
	  score += * (MATENTRY_T *) (m + *s++);
	}
	if (score > threshold) {
	  significance = G[score - min_score];
	  if (Num_positives >= MAXPOSITIVES) {
	    fprintf(stderr, "Error! Too many positives: Limit is %d\n",
		    MAXPOSITIVES);
	    exit(1);
	  } else {
	    hit_translate(hit,Seq,start,width);
	    Entries[Num_positives++] =
	      Entry_new(Name,start,width,significance,hit);
	  }
	}
      }
    }
  }
  free(hit);
}
#else
static void
score_sequences (FILE *seq_fp, FILE *desc_fp, int *score_matrix, 
		 int threshold, int *sfloor, double *G, int min_score, 
		 int width) {
  register int score;
  register int i;
  int *sfloor_ptr;
  int length, start, end, pos, descript_len;
  double significance;
  char *hit = (char *) calloc(width+1,sizeof(char)), *description;
  short *sequence_int = NULL, *s;
  long descript_ptr, last_descript_ptr = 0L;
  bool newseq;

  description = (char *) calloc(MAXNAMELEN+1,sizeof(char));
  Num_positives = 0;

  while (!feof(seq_fp)) {
    fread(&length,sizeof(int),1,seq_fp);
    fread(&descript_ptr,sizeof(long),1,seq_fp);
    fread(&descript_len,sizeof(int),1,seq_fp);
    sequence_int = (short *) calloc(length,sizeof(short));
    fread(&sequence_int[0],sizeof(short),length,seq_fp);
    newseq = TRUE;

    if (length >= width) {
      end = length - width;
      for (start = 0; start <= end; ++start) {
	s = sequence_int + start;
	score = score_matrix[*s++];
	for (i = SCOREMAT_SKIP, pos = 1, sfloor_ptr = sfloor;
	     (pos < width) && (score >= *sfloor_ptr++);
	     i += SCOREMAT_SKIP, ++pos) {
	  score += score_matrix[i + *s++];
	}
	if (score > threshold) {
	  significance = G[score - min_score];
	  if (Num_positives >= MAXPOSITIVES) {
	    fprintf(stderr, "Error! Too many positives: Limit is %d\n",
		    MAXPOSITIVES);
	    exit(1);
	  } else {
	    hit_translate(hit,sequence_int,start,width);
	    if (newseq) {
	      fseek(desc_fp,descript_ptr-last_descript_ptr,SEEK_CUR);
	      if (descript_len > MAXNAMELEN) {
		fread(&description[0],sizeof(char),MAXNAMELEN,desc_fp);
		description[MAXNAMELEN] = '\0';
		last_descript_ptr = descript_ptr + (long) MAXNAMELEN;
	      } else {
		fread(&description[0],sizeof(char),descript_len,desc_fp);
		description[descript_len] = '\0';
		last_descript_ptr = descript_ptr + (long) descript_len;
	      }
	      newseq = FALSE;
	    }
	    Entries[Num_positives++] =
	      Entry_new(description,start,width,significance,hit);
	  }
	}
      }
    }

    free(sequence_int);
  }

  free(hit);
  free(description);
}
#endif      

void
print_entries () {
  int i;
  
  for (i = 0; i < Num_positives; ++i) {
    Entry_print(Entries[i]);
  }
}

static void
parse_arguments (char **matrixfile, char **directory, char **filebase, 
		 double *e_threshold, int *s_threshold,
		 int argc, char *argv[]) {
  ++argv; --argc;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if ((*argv)[1] == 'e') {
	*e_threshold = atof(argv[1]); argc -= 2; ++argv; ++argv;
      } else if ((*argv)[1] == 's') {
	*s_threshold = atoi(argv[1]); argc -= 2; ++argv; ++argv;
      } else if ((*argv)[1] == 'd') {
	*directory = *(argv+1); argc -= 2; ++argv; ++argv;
      } else if ((*argv)[1] == 'f') {
	*filebase = *(argv+1); argc -= 2; ++argv; ++argv;
      } else {
	fprintf(stderr, "I don't understand.\n");
	exit(1);
      }
    } else {
      *matrixfile = *argv; argc -= 1; argv += 1;
    }
  }
}

int
main (int argc, char *argv[]) {
  FILE *fp, *seq_fp, *desc_fp;
  int i = 0, j;
  char *matrixfile = NULL, *directory = DATA_DIR, 
    *filebase = DEFAULT_SCAN_FILE, *seqfile = NULL, *descfile = NULL,
    accession[MAXNAMELEN], description[MAXNAMELEN];
  int baselength = 0;
  int **calib_matrix, *score_matrix, *min_scores, *max_scores, *sfloor;
  int width, min_score, max_score, threshold, s_threshold = INFINITY;
  unsigned long N;
  double *G, e_threshold = DEFAULT_SCAN_THRESHOLD, p_threshold;

  parse_arguments(&matrixfile,&directory,&filebase,
		  &e_threshold,&s_threshold,argc,argv);
  calib_matrix = Intmatrix_new(MAXWIDTH,NAT_AAS);
  score_matrix = (int *) calloc(MAXWIDTH*(NAT_AAS+1),sizeof(int));

  if (matrixfile == NULL) {
    fp = stdin;
  } else {
    if (!(fp = fopen(matrixfile,"r"))) {
      fprintf(stderr, "Can't open file %s\n",matrixfile);
      exit(1);
    }
  }

  baselength = strlen(directory);
  
  if (directory[strlen(directory)-1] != '/') {
    baselength += 1;
  }
  
  baselength += strlen(filebase);
  baselength += strlen(".seq");
  
  seqfile = (char *) calloc(baselength+1,sizeof(char));
  descfile = (char *) calloc(baselength+1,sizeof(char));
  
  if (strchr(filebase,'/')) {
    strcpy(seqfile,"");
    strcpy(descfile,"");
  } else {
    strcpy(seqfile,directory);
    strcpy(descfile,directory);
    
    if (directory[strlen(directory)-1] != '/') {
      strcat(seqfile,"/");
      strcat(descfile,"/");
    }
  }
  
  strcat(seqfile,filebase);
  strcat(descfile,filebase);
  strcat(seqfile,".seq");
  strcat(descfile,".des");

  while ((width = read_matrices(fp, calib_matrix, score_matrix,
				accession, description, &N)) > 0) {
    fprintf(stderr, "%d\r", i++);
    p_threshold = e_threshold / N;

    min_scores = Intvector_new(width);
    max_scores = Intvector_new(width);

    CDF_bounds(&min_score,&max_score,min_scores,max_scores,
	       calib_matrix,width);
    
    Intvector_free(&min_scores);
    Intvector_free(&max_scores);
    
    G = (double *) calloc(max_score-min_score+1,sizeof(double));
    CDF_compute(G,calib_matrix,width,min_score,max_score);
    
    threshold = CDF_getThreshold(G,p_threshold,min_score,max_score);

    if ((s_threshold != INFINITY) && (s_threshold > threshold)) {
      threshold = s_threshold;
    }

    if (threshold != max_score) {
      --threshold;
      sfloor = Intvector_new(width);
      sfloor_compute(sfloor,calib_matrix,width,threshold);

#if TEXT
      if (!(fp = fopen(seqfile,"r"))) {
	fprintf(stderr, "Can't open file %s\n",seqfile);
	exit(1);
      }

      score_sequences(fp,score_matrix,threshold,sfloor,G,min_score,width);
      fclose(fp);
#else 
      if (!(seq_fp = fopen(seqfile,"rb"))) {
	fprintf(stderr, "Can't open file %s\n",seqfile);
	exit(1);
      }
      
      if (!(desc_fp = fopen(descfile,"rb"))) {
	fprintf(stderr, "Can't open file %s\n",descfile);
	exit(1);
      }

      score_sequences(seq_fp,desc_fp,score_matrix,threshold,sfloor,
		      G,min_score,width);
      free (sfloor);
      fclose(desc_fp);
      fclose(seq_fp);
#endif
      if (Num_positives) {
	qsort(Entries,Num_positives,sizeof(Entry_T),Entry_cmp);
	fprintf(stdout, ">%s|%s\n", accession, description);
	print_entries();

	for (j = 0; j < Num_positives; j++) {
	  if (Entries[j] != NULL) {
	    Entry_free(&Entries[j]);
	    Entries[j] = NULL;
	  }
	}
      }
    }
    
    free (G);
  }

  free (descfile);
  free (seqfile);

  free (score_matrix);
  Intmatrix_free(&calib_matrix, MAXWIDTH);

  fclose(fp);

  fclose (stdout);
  fclose (stderr);

  return (0);
}
